import json
import requests

from typing import Optional, Union


class AbstractAttribute:
    def __get__(self, obj, type):
        raise NotImplementedError("This attribute must be set")


class BaseClient:
    """
    Base client
    """
    baseurl: Union[AbstractAttribute, str] = AbstractAttribute()

    def __init__(self) -> None:
        self.s = requests.Session()

    def _get(self, request: str, params: Optional[dict] = None) -> dict:
        resp = self.s.get(self.baseurl + request, params=params)
        resp.raise_for_status()

        return json.loads(json.dumps((resp.json())))



import json
from wtfis.models.virustotal import (
    Domain,
    HistoricalWhois,
    Resolutions,
)


class TestVirustotalModels:
    def test_domain_1(self, test_data):
        domain = Domain.parse_obj(json.loads(test_data("vt_domain_gist.json")))

        assert domain.data.id_ == "gist.github.com"
        assert domain.data.type_ == "domain"
        assert domain.data.attributes.last_analysis_stats.malicious == 0
        assert domain.data.attributes.last_analysis_stats.suspicious == 0
        assert domain.data.attributes.last_analysis_stats.harmless == 84
        assert domain.data.attributes.last_analysis_stats.undetected == 10
        assert list(domain.data.attributes.last_analysis_results.__root__.keys())[:2] == [
            "CMC Threat Intelligence",
            "Snort IP sample list",
        ]

    def test_resolutions_1(self, test_data):
        res = Resolutions.parse_obj(json.loads(test_data("vt_resolutions_gist.json")))

        assert res.meta.count == 37
        assert len(res.data) == 10
        assert res.data[1].attributes.ip_address == "192.30.255.113"
        assert res.data[1].attributes.ip_address_last_analysis_stats.malicious == 1
        assert res.data[1].attributes.ip_address_last_analysis_stats.suspicious == 0
        assert res.data[1].attributes.ip_address_last_analysis_stats.harmless == 82
        assert res.data[1].attributes.ip_address_last_analysis_stats.undetected == 11
        assert res.data[1].attributes.date == 1655835054

    def test_historical_whois_1(self, test_data):
        whois = HistoricalWhois.parse_obj(json.loads(test_data("vt_whois_bbc.json")))

        assert whois.meta.count == 5
        assert whois.data[0].attributes.whois_map.domain == ""
        assert whois.data[0].attributes.whois_map.name_servers == ["dns1.bbc.com"]
        assert whois.data[0].attributes.whois_map.registrant_org is None

    def test_historical_whois_2(self, test_data):
        whois = HistoricalWhois.parse_obj(json.loads(test_data("vt_whois_example.json")))

        assert whois.meta.count == 9
        assert whois.data[0].attributes.whois_map.domain == "example.com"
        assert whois.data[0].attributes.whois_map.registrant_name == "1f8f4166599d23ee"
        assert whois.data[0].attributes.whois_map.registrant_org == "b46a98a26fe2fd9f"
        assert whois.data[0].attributes.whois_map.name_servers == [
            "lochlan.ns.cloudflare.com",
            "nelci.ns.cloudflare.com",
        ]
        assert whois.data[6].attributes.whois_map is None
        assert whois.data[6].attributes.first_seen_date == 1574839630



import pytest
from pathlib import Path


def open_test_data(fname: str) -> str:
    path = Path(__file__).parent.resolve() / "test_data" / fname
    with open(path) as f:
        return f.read()


@pytest.fixture(scope="module")
def test_data():
    return open_test_data



import pytest

from freezegun import freeze_time
from rich.text import Text

from wtfis.utils import (
    iso_date,
    older_than,
    smart_join,
    error_and_exit,
)


class TestUtils:
    def test_iso_date_1(self):
        assert iso_date(1660428690) == "2022-08-13T22:11:30Z"

    def test_iso_date_2(self):
        assert iso_date("2017-04-26T15:43:49.000-07:00") == "2017-04-26T22:43:49Z"

    def test_iso_date_3(self):
        assert iso_date(None) is None

    @freeze_time("2022-08-12")
    def test_older_than(self):
        assert older_than(1659146024, 30) is False
        assert older_than(1657418024, 30) is True

    def test_smart_join_1(self):
        output = smart_join(*["foo", "bar", "baz"])
        assert type(output) is Text
        assert str(output) == "foo, bar, baz"
        assert output == (
            Text()
            .append("foo")
            .append(", ", style="default")
            .append("bar")
            .append(", ", style="default")
            .append("baz")
        )

    def test_smart_join_2(self):
        output = smart_join(*["foo", "bar", "baz"], style="blue")
        assert type(output) is Text
        assert str(output) == "foo, bar, baz"
        assert output == (
            Text()
            .append("foo", style="blue")
            .append(", ", style="default")
            .append("bar", style="blue")
            .append(", ", style="default")
            .append("baz", style="blue")
        )

    def test_smart_join_3(self):
        output_1 = smart_join(*[])
        assert type(output_1) is str
        assert output_1 == ""

        output_2 = smart_join(None)
        assert type(output_2) is str
        assert output_2 == ""

    def test_error_and_exit(self):
        with pytest.raises(SystemExit) as e:
            error_and_exit("foo", 2)
        assert e.value.code == 2



import sys

from datetime import datetime, timedelta, timezone
from typing import Optional, Union

from rich.text import Text


def iso_date(ts: Optional[Union[str, int]]) -> Optional[str]:
    """ Convert any time to a standard format """
    std_utc = "%Y-%m-%dT%H:%M:%SZ"
    if isinstance(ts, int):
        return datetime.fromtimestamp(ts, tz=timezone.utc).strftime(std_utc)
    elif isinstance(ts, str):
        return datetime.fromisoformat(ts).astimezone(timezone.utc).strftime(std_utc)
    return None


def older_than(ts: int, days: int) -> bool:
    """ Tells if a timestamp is older than X days. "ts" must be epoch format """
    return datetime.fromtimestamp(ts) < datetime.now() - timedelta(days=days)


def smart_join(*items: Optional[str], style: Optional[str] = None) -> Union[Text, str]:
    text = Text()
    for idx, item in enumerate(items):
        if item:
            text.append(item, style=style)
            if idx < len(items) - 1:
                text.append(", ", style="default")
    return text if str(text) != "" else ""


def error_and_exit(message: str, status: int = 1):
    print(message, file=sys.stderr)
    raise SystemExit(status)



from typing import Optional

from wtfis.clients.base import BaseClient
from wtfis.models.virustotal import Domain, HistoricalWhois, Resolutions


class VTClient(BaseClient):
    """
    Virustotal client
    """
    baseurl = "https://www.virustotal.com/api/v3"

    def __init__(self, api_key: str) -> None:
        super().__init__()
        self.s.headers.update({
            "x-apikey": api_key,
            "Accept": "application/json",
        })

    def get_domain(self, domain: str) -> Domain:
        return Domain.parse_obj(self._get(f"/domains/{domain}"))

    def get_domain_resolutions(self, domain: str) -> Optional[Resolutions]:
        return Resolutions.parse_obj(self._get(f"/domains/{domain}/resolutions"))

    def get_domain_whois(self, domain: str) -> HistoricalWhois:
        return HistoricalWhois.parse_obj(self._get(f"/domains/{domain}/historical_whois"))



from rich.columns import Columns
from rich.console import (
    Console,
    Group,
    RenderableType,
    group,
)
from rich.padding import Padding
from rich.panel import Panel
from rich.table import Table
from rich.text import Text
from typing import Generator, List, Optional, Union

from wtfis.models.ipwhois import IpWhois
from wtfis.models.passivetotal import Whois
from wtfis.models.virustotal import (
    Domain,
    HistoricalWhois,
    LastAnalysisStats,
    PopularityRanks,
    Resolutions,
)
from wtfis.ui.theme import Theme
from wtfis.utils import iso_date, older_than, smart_join


class View:
    """
    Handles the look of the output
    """
    vt_gui_baseurl_domain = "https://virustotal.com/gui/domain"
    vt_gui_baseurl_ip = "https://virustotal.com/gui/ip-address"
    pt_gui_baseurl = "https://community.riskiq.com/search"

    def __init__(
        self,
        console: Console,
        domain: Domain,
        resolutions: Optional[Resolutions],
        whois: Union[Whois, HistoricalWhois],
        ip_enrich: List[IpWhois] = [],
        max_resolutions: int = 3,
    ) -> None:
        self.console = console
        self.domain = domain
        self.resolutions = resolutions
        self.whois = whois
        self.ip_enrich = ip_enrich
        self.max_resolutions = max_resolutions
        self.theme = Theme()

    def _vendors_who_flagged_malicious(self) -> List[str]:
        vendors = []
        for key, result in self.domain.data.attributes.last_analysis_results.__root__.items():
            if result.category == "malicious":
                vendors.append(key)
        return vendors

    def _gen_heading_text(self, heading: str, hyperlink: Optional[str] = None) -> Text:
        text = Text(justify="center", end="\n")
        style = f"{self.theme.heading} link {hyperlink}" if hyperlink else self.theme.heading
        return text.append(heading, style=style)

    def _gen_table(self, *params) -> Union[Table, str]:
        """ Each param should be a tuple of (field, value) """
        # Set up table
        grid = Table.grid(expand=False, padding=(0, 1))
        grid.add_column(style=self.theme.table_field)                # Field
        grid.add_column(style=self.theme.table_value, max_width=38)  # Value

        # Populate rows
        valid_rows = 0
        for item in params:
            field, value = item
            if value is None or str(value) == "":  # Skip if no value
                continue
            grid.add_row(field, value)
            valid_rows += 1

        # Return None if no rows generated
        return grid if valid_rows > 0 else ""

    @group()
    def _gen_group(self, content: List[RenderableType]) -> Generator:
        for item in content:
            yield item

    def _gen_info(self, body: RenderableType, heading: Optional[Text] = None) -> RenderableType:
        return Group(heading, body) if heading else body

    def _gen_panel(self, title: str, renderable: RenderableType) -> Panel:
        panel_title = Text(title, style=self.theme.panel_title)
        return Panel(renderable, title=panel_title, expand=False)

    def _gen_vt_analysis_stats(
        self,
        stats: LastAnalysisStats,
        vendors: Optional[List[str]] = None
    ) -> Text:
        # Custom style
        stats_style = self.theme.error if stats.malicious >= 1 else self.theme.info

        # Total count
        total = stats.harmless + stats.malicious + stats.suspicious + stats.timeout + stats.undetected

        # Text
        text = Text(f"{stats.malicious}/{total} malicious", style=stats_style)

        # Include list of vendors that flagged malicious
        if vendors:
            text.append("\n")
            text.append(smart_join(*vendors, style=self.theme.vendor_list))

        return text

    def _gen_vt_reputation(self, reputation: int) -> Text:
        # Custom style
        def rep_style(reputation: int) -> str:
            if reputation > 0:
                return self.theme.info
            elif reputation < 0:
                return self.theme.error
            return "default"

        return Text(str(reputation), style=rep_style(reputation))

    def _gen_vt_popularity(self, popularity_ranks: PopularityRanks) -> Optional[Text]:
        if len(popularity_ranks.__root__) == 0:
            return None

        text = Text()
        for source, popularity in popularity_ranks.__root__.items():
            text.append(f"{source} (")
            text.append(str(popularity.rank), style=self.theme.inline_stat)
            text.append(")")
            if source != list(popularity_ranks.__root__.keys())[-1]:
                text.append("\n")
        return text

    def _get_ip_enrichment(self, ip: str) -> Optional[IpWhois]:
        for ipwhois in self.ip_enrich:
            if ipwhois.ip == ip:
                return ipwhois
        return None

    def whois_panel(self) -> Optional[Panel]:
        # Using PT Whois
        if isinstance(self.whois, Whois):
            heading = self._gen_heading_text(
                self.whois.domain,
                hyperlink=f"{self.pt_gui_baseurl}/{self.whois.domain}/whois"
            )
            body = self._gen_table(
                ("Registrar:", self.whois.registrar),
                ("Organization:", self.whois.organization),
                ("Name:", self.whois.name),
                ("Email:", self.whois.contactEmail),
                ("Phone:", self.whois.registrant.telephone),
                ("Street:", self.whois.registrant.street),
                ("City:", self.whois.registrant.city),
                ("State:", self.whois.registrant.state),
                ("Country:", self.whois.registrant.country),
                ("Nameservers:", smart_join(*self.whois.nameServers, style=self.theme.nameserver_list)),
                ("Registered:", iso_date(self.whois.registered)),
                ("Updated:", iso_date(self.whois.registryUpdatedAt)),
                ("Expires:", iso_date(self.whois.expiresAt)),
            )
        # Using VT HistoricalWhois
        else:
            if not self.whois.data:
                return None

            # Use the first, i.e. latest whois entry
            attribs = self.whois.data[0].attributes

            # Check for empty whois map
            if not attribs.whois_map:
                return self._gen_panel("whois", Text("Unable to gather whois data", style=self.theme.disclaimer))

            # Admin location
            admin_location = smart_join(
                attribs.whois_map.admin_city,
                attribs.whois_map.admin_state,
                attribs.whois_map.admin_country,
            )

            # Name servers
            name_servers = attribs.whois_map.name_servers if attribs.whois_map.name_servers else []

            heading = self._gen_heading_text(attribs.whois_map.domain)
            body = self._gen_table(
                ("Registrar:", attribs.whois_map.registrar),
                ("Organization:", attribs.whois_map.registrant_org),
                ("Name:", attribs.whois_map.registrant_name),
                ("Email:", attribs.whois_map.registrant_email),
                ("Country:", attribs.registrant_country),
                ("Admin Location:", admin_location),
                ("Nameservers:", smart_join(*name_servers, style=self.theme.nameserver_list)),
                ("Registered:", attribs.whois_map.creation_date or attribs.whois_map.registered_on),
                ("Updated:", attribs.whois_map.updated_date or attribs.whois_map.last_updated),
                ("Expires:", attribs.whois_map.expiry_date or attribs.whois_map.expiry_date_alt),
            )

        # Return None if body is empty
        return self._gen_panel("whois", self._gen_info(body, heading)) if body else None

    def domain_panel(self) -> Panel:
        attributes = self.domain.data.attributes

        # Analysis
        analysis = self._gen_vt_analysis_stats(
            attributes.last_analysis_stats,
            self._vendors_who_flagged_malicious()
        )

        # Reputation
        reputation = self._gen_vt_reputation(attributes.reputation)

        # Popularity
        popularity = self._gen_vt_popularity(attributes.popularity_ranks)

        # Content
        heading = self._gen_heading_text(
            self.domain.data.id_,
            hyperlink=f"{self.vt_gui_baseurl_domain}/{self.domain.data.id_}"
        )
        body = self._gen_table(
            ("Analysis:", analysis),
            ("Reputation:", reputation),
            ("Popularity:", popularity),
            ("Last Modified:", iso_date(attributes.last_modification_date)),
            ("Last Seen:", iso_date(attributes.last_dns_records_date)),
        )
        return self._gen_panel("virustotal", self._gen_info(body, heading))

    def resolutions_panel(self) -> Optional[Panel]:
        # Skip if no resolutions data
        if not self.resolutions:
            return None

        content = []
        for idx, ip in enumerate(self.resolutions.data):
            if idx == self.max_resolutions:
                break
            attributes = ip.attributes

            # Analysis
            analysis = self._gen_vt_analysis_stats(attributes.ip_address_last_analysis_stats)

            # IP Enrichment
            enrich = self._get_ip_enrichment(attributes.ip_address)

            # Content
            heading = self._gen_heading_text(
                attributes.ip_address,
                hyperlink=f"{self.vt_gui_baseurl_ip}/{attributes.ip_address}"
            )
            data = [
                ("Analysis:", analysis),
                ("Resolved:", iso_date(attributes.date)),
            ]
            if enrich:
                data += [
                    ("ASN:", f"{enrich.connection.asn} ({enrich.connection.org})"),
                    ("ISP:", enrich.connection.isp),
                    ("Location:", smart_join(enrich.city, enrich.region, enrich.country)),
                ]

            # Include a disclaimer if last seen is older than 1 year
            if older_than(attributes.date, 365):
                body = Group(
                    self._gen_table(*data),
                    Text("**Enrichment data may be inaccurate", style=self.theme.disclaimer),
                )  # type: Union[Group, Table, str]
            else:
                body = self._gen_table(*data)

            content.append(self._gen_info(body, heading))

            # Add extra line break if not last item in list
            if (
                idx < self.max_resolutions - 1 and
                idx < len(self.resolutions.data) - 1
            ):
                content.append("")

        # Info about how many more IPs were not shown
        if self.max_resolutions < self.resolutions.meta.count:
            content.append(
                Text(
                    f"\n+{self.resolutions.meta.count - self.max_resolutions} more",
                    justify="center",
                    style=self.theme.footer,
                )
            )

        # Render results, if existent
        if content:
            return self._gen_panel("resolutions", self._gen_group(content))

        # No result
        return None

    def print(self, one_column: bool = False) -> None:
        renderables = [i for i in (
            self.domain_panel(),
            self.resolutions_panel(),
            self.whois_panel(),
        ) if i is not None]

        if one_column:
            self.console.print(Group(*([""] + renderables)))  # type: ignore
        else:
            self.console.print(Padding(Columns(renderables), (1, 0)))



import argparse
import os

from dotenv import load_dotenv
from pathlib import Path
from pydantic import ValidationError
from requests.exceptions import HTTPError, JSONDecodeError
from rich.console import Console

from wtfis.clients.ipwhois import IpWhoisClient
from wtfis.clients.passivetotal import PTClient
from wtfis.clients.virustotal import VTClient
from wtfis.utils import error_and_exit
from wtfis.ui.progress import get_progress
from wtfis.ui.view import View
from wtfis.version import get_version


def parse_env():
    DEFAULT_ENV_FILE = Path().home() / ".env.wtfis"

    # Load the file
    load_dotenv(DEFAULT_ENV_FILE)

    # Exit if required environment variables don't exist
    for envvar in (
        "VT_API_KEY",
    ):
        if not os.environ.get(envvar):
            error = f"Error: Environment variable {envvar} not set"
            if not DEFAULT_ENV_FILE.exists():
                error = error + f"\nEnv file {DEFAULT_ENV_FILE} was not found either. Did you forget?"
            error_and_exit(error)


def parse_args():
    DEFAULT_MAX_RESOLUTIONS = 3

    parser = argparse.ArgumentParser()
    parser.add_argument("hostname", help="Hostname or domain")
    parser.add_argument(
        "-m", "--max-resolutions", metavar="N",
        help=f"Maximum number of resolutions to show (default: {DEFAULT_MAX_RESOLUTIONS})",
        type=int,
        default=DEFAULT_MAX_RESOLUTIONS
    )
    parser.add_argument("-n", "--no-color", help="Show output without colors", action="store_true")
    parser.add_argument("-1", "--one-column", help="Display results in one column", action="store_true")
    parser.add_argument(
        "-V", "--version",
        help="Print version number",
        action="version",
        version=get_version()
    )

    # Validation
    parsed = parser.parse_args()
    if parsed.max_resolutions > 10:
        argparse.ArgumentParser().error("Maximum --max-resolutions value is 10")

    return parsed


def main():
    # Load environment variables
    parse_env()

    # Args
    args = parse_args()

    # Instantiate the console
    console = Console(no_color=True) if args.no_color else Console()

    # Progress animation controller
    progress = get_progress(console)

    # Fetch data
    with progress:
        try:
            # Virustotal domain
            task1 = progress.add_task("Fetching data from Virustotal")
            vt = VTClient(os.environ.get("VT_API_KEY"))
            progress.update(task1, advance=33)
            domain = vt.get_domain(args.hostname)
            progress.update(task1, advance=33)

            # Resolutions and IP enrichments
            if args.max_resolutions != 0:
                resolutions = vt.get_domain_resolutions(args.hostname)
                progress.update(task1, advance=33)

                task2 = progress.add_task("Fetching IP enrichments from IPWhois")
                ipwhois = IpWhoisClient()
                progress.update(task2, advance=50)
                ip_enrich = ipwhois.bulk_get_ipwhois(resolutions, args.max_resolutions)
                progress.update(task2, advance=50)
            else:
                resolutions = None
                ip_enrich = []
            progress.update(task1, completed=100)

            # Whois
            # Use Passivetotal if relevant environment variables exist, otherwise keep using VT
            if os.environ.get("PT_API_USER") and os.environ.get("PT_API_KEY"):
                task3 = progress.add_task("Fetching domain whois from Passivetotal")
                pt = PTClient(os.environ.get("PT_API_USER"), os.environ.get("PT_API_KEY"))
                progress.update(task3, advance=50)
                whois = pt.get_whois(args.hostname)
                progress.update(task3, advance=50)
            else:
                task3 = progress.add_task("Fetching domain whois from Virustotal")
                whois = vt.get_domain_whois(args.hostname)
                progress.update(task3, advance=100)
        except (HTTPError, JSONDecodeError) as e:
            progress.stop()
            error_and_exit(f"Error fetching data: {e}")
        except ValidationError as e:
            progress.stop()
            error_and_exit(f"Data model validation error: {e}")

    # Output
    view = View(
        console,
        domain,
        resolutions,
        whois,
        ip_enrich,
        max_resolutions=args.max_resolutions,
    )
    view.print(one_column=args.one_column)



import pytest
import json

from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.text import Span, Text
from unittest.mock import MagicMock

from wtfis.models.ipwhois import IpWhois
from wtfis.models.passivetotal import Whois
from wtfis.models.virustotal import (
    Domain,
    HistoricalWhois,
    Resolutions,
)
from wtfis.ui.view import View


@pytest.fixture()
def view01(test_data):
    """ gist.github.com with PT whois. Complete test of all panels. """
    return View(
        console=Console(),
        domain=Domain.parse_obj(json.loads(test_data("vt_domain_gist.json"))),
        resolutions=Resolutions.parse_obj(json.loads(test_data("vt_resolutions_gist.json"))),
        whois=Whois.parse_obj(json.loads(test_data("pt_whois_gist.json"))),
        ip_enrich=[IpWhois.parse_obj(o) for o in json.loads(test_data("ipwhois_gist.json"))],
    )


@pytest.fixture()
def view02(test_data):
    """
    gist.github.com VT whois. Resolution and whois tests only. Test empty enrichment
    and max_resolutions=1
    """
    return View(
        console=Console(),
        domain=MagicMock(),
        resolutions=Resolutions.parse_obj(json.loads(test_data("vt_resolutions_gist.json"))),
        whois=HistoricalWhois.parse_obj(json.loads(test_data("vt_whois_gist.json"))),
        ip_enrich=[],
        max_resolutions=1,
    )


@pytest.fixture()
def view03(test_data):
    """ bbc.co.uk VT whois. Whois panel test only. Test whois with no domain field. """
    return View(
        console=Console(),
        domain=MagicMock(),
        resolutions=MagicMock(),
        whois=HistoricalWhois.parse_obj(json.loads(test_data("vt_whois_bbc.json"))),
        ip_enrich=[],
    )


@pytest.fixture()
def view04(test_data):
    """
    google.com domain. Domain and resolution test only. Test domain with 1 malicious
    analysis point, and empty resolutions.
    """
    return View(
        console=Console(),
        domain=Domain.parse_obj(json.loads(test_data("vt_domain_google.json"))),
        resolutions=None,
        whois=MagicMock(),
        ip_enrich=[],
    )


@pytest.fixture()
def view05(test_data):
    """ tucows.com domain. Domain test only. Test domain with negative reputation and no popularity."""
    return View(
        console=Console(),
        domain=Domain.parse_obj(json.loads(test_data("vt_domain_tucows.json"))),
        resolutions=MagicMock(),
        whois=MagicMock(),
        ip_enrich=[],
    )


@pytest.fixture()
def view06(test_data):
    """ exmple.com VT whois. Whois test only. Test empty whois_map."""
    return View(
        console=Console(),
        domain=MagicMock(),
        resolutions=MagicMock(),
        whois=HistoricalWhois.parse_obj(json.loads(test_data("vt_whois_example_2.json"))),
        ip_enrich=[],
    )


class TestView01:
    def test_domain_panel(self, view01):
        domain = view01.domain_panel()
        assert type(domain) is Panel
        assert domain.title == Text("virustotal")

        # Heading
        assert domain.renderable.renderables[0] == Text(
            "gist.github.com",
            spans=[Span(0, 15, 'bold yellow link https://virustotal.com/gui/domain/gist.github.com')]
        )

        # Table
        table = domain.renderable.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == "bold bright_magenta"
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "Analysis:",
            "Reputation:",
            "Popularity:",
            "Last Modified:",
            "Last Seen:",
        ]
        assert table.columns[1].style == "none"
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text("0/94 malicious"),
            Text("0"),
            Text(
                "Majestic (532)\nCisco Umbrella (39463)",
                spans=[
                    Span(10, 13, "cyan"),
                    Span(31, 36, "cyan"),
                ]
            ),
            "2022-08-16T06:14:59Z",
            "2022-08-15T22:25:30Z",
        ]

    def test_resolutions_panel(self, view01):
        res = view01.resolutions_panel()
        assert type(res) is Panel

        # Entry 1
        group = res.renderable.renderables[0].renderables

        # Heading
        assert group[0] == Text(
            "13.234.210.38",
            spans=[Span(0, 13, "bold yellow link https://virustotal.com/gui/ip-address/13.234.210.38")],
        )

        # Table
        assert group[1].columns[0].style == "bold bright_magenta"
        assert group[1].columns[0].justify == "left"
        assert group[1].columns[0]._cells == [
            "Analysis:",
            "Resolved:",
            "ASN:",
            "ISP:",
            "Location:",
        ]
        assert group[1].columns[1].style == "none"
        assert group[1].columns[1].justify == "left"
        assert group[1].columns[1]._cells == [
            Text("0/94 malicious"),
            "2022-08-06T14:56:20Z",
            "16509 (Amazon Data Services India)",
            "Amazon.com, Inc.",
            Text(
                "Mumbai, Maharashtra, India",
                spans=[
                    Span(6, 8, "default"),
                    Span(19, 21, "default"),
                ]
            )
        ]

        # Spacing
        assert res.renderable.renderables[1] == ""

        # Entry 2
        group = res.renderable.renderables[2].renderables

        # Heading
        assert group[0] == Text(
            "192.30.255.113",
            spans=[Span(0, 14, "bold yellow link https://virustotal.com/gui/ip-address/192.30.255.113")],
        )

        # Table
        assert group[1].columns[0].style == "bold bright_magenta"
        assert group[1].columns[0].justify == "left"
        assert group[1].columns[0]._cells == [
            "Analysis:",
            "Resolved:",
            "ASN:",
            "ISP:",
            "Location:",
        ]
        assert group[1].columns[1].style == "none"
        assert group[1].columns[1].justify == "left"
        assert group[1].columns[1]._cells == [
            Text("1/94 malicious"),
            "2022-06-21T18:10:54Z",
            "36459 (GitHub, Inc.)",
            "GitHub, Inc.",
            Text(
                "Seattle, Washington, United States",
                spans=[
                    Span(7, 9, "default"),
                    Span(19, 21, "default"),
                ]
            )
        ]

        # Spacing
        assert res.renderable.renderables[3] == ""

        # Entry 3 (NOTE: Timestamp on data modified to be really old)
        group = res.renderable.renderables[4].renderables

        # Heading
        assert group[0] == Text(
            "13.234.176.102",
            spans=[Span(0, 14, "bold yellow link https://virustotal.com/gui/ip-address/13.234.176.102")],
        )

        # Unlike the previous entries, the table is inside a group of (Table, Text) due to
        # old timestamp warning
        table = group[1].renderables[0]
        assert table.columns[0].style == "bold bright_magenta"
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "Analysis:",
            "Resolved:",
            "ASN:",
            "ISP:",
            "Location:",
        ]
        assert table.columns[1].style == "none"
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text("0/94 malicious"),
            "2015-08-17T07:11:53Z",
            "16509 (Amazon Data Services India)",
            "Amazon.com, Inc.",
            Text(
                "Mumbai, Maharashtra, India",
                spans=[
                    Span(6, 8, "default"),
                    Span(19, 21, "default"),
                ]
            )
        ]

        # Old timestamp warning
        assert group[1].renderables[1] == Text("**Enrichment data may be inaccurate")

        # Spacing and remaining count
        assert res.renderable.renderables[5] == Text("\n+34 more")

    def test_whois_panel(self, view01):
        whois = view01.whois_panel()
        assert type(whois) is Panel
        assert whois.title == Text("whois")

        # Heading
        assert whois.renderable.renderables[0] == Text(
            "github.com",
            spans=[Span(0, 10, 'bold yellow link https://community.riskiq.com/search/github.com/whois')]
        )

        # Table
        table = whois.renderable.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == "bold bright_magenta"
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "Registrar:",
            "Organization:",
            "Name:",
            "Email:",
            "State:",
            "Country:",
            "Nameservers:",
            "Registered:",
            "Updated:",
            "Expires:",
        ]
        assert table.columns[1].style == "none"
        assert table.columns[1].justify == "left"
        assert [str(c) for c in table.columns[1]._cells] == [
            "MarkMonitor Inc.",
            "GitHub, Inc.",
            "N/A",
            "abusecomplaints@markmonitor.com",
            "CA",
            "US",
            ("dns1.p08.nsone.net, dns2.p08.nsone.net, dns3.p08.nsone.net, dns4.p08.nsone.net, ns-1283.awsdns-32.org, "
             "ns-1707.awsdns-21.co.uk, ns-421.awsdns-52.com, ns-520.awsdns-01.net"),
            "2007-10-09T18:20:50Z",
            "2020-09-08T09:18:27Z",
            "2022-10-09T07:00:00Z"
        ]


class TestView02:
    def test_resolutions_panel(self, view02):
        res = view02.resolutions_panel()
        assert type(res) is Panel

        # Entry 1
        group = res.renderable.renderables[0].renderables

        # Heading
        assert group[0] == Text(
            "13.234.210.38",
            spans=[Span(0, 13, "bold yellow link https://virustotal.com/gui/ip-address/13.234.210.38")],
        )

        # Table
        assert group[1].columns[0].style == "bold bright_magenta"
        assert group[1].columns[0].justify == "left"
        assert group[1].columns[0]._cells == [
            "Analysis:",
            "Resolved:",
        ]
        assert group[1].columns[1].style == "none"
        assert group[1].columns[1].justify == "left"
        assert group[1].columns[1]._cells == [
            Text("0/94 malicious"),
            "2022-08-06T14:56:20Z",
        ]

        # Spacing and remaining count
        assert res.renderable.renderables[1] == Text("\n+36 more")

    def test_whois_panel(self, view02):
        whois = view02.whois_panel()
        assert type(whois) is Panel
        assert whois.title == Text("whois")

        # Heading
        assert whois.renderable.renderables[0] == Text(
            "GITHUB.COM",
            spans=[Span(0, 10, 'bold yellow')]
        )

        # Table
        table = whois.renderable.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == "bold bright_magenta"
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "Registrar:",
            "Nameservers:",
            "Registered:",
            "Updated:",
            "Expires:",
        ]
        assert table.columns[1].style == "none"
        assert table.columns[1].justify == "left"
        assert [str(c) for c in table.columns[1]._cells] == [
            "MarkMonitor Inc.",
            "DNS1.P08.NSONE.NET, DNS2.P08.NSONE.NET, DNS3.P08.NSONE.NET, "
            "DNS4.P08.NSONE.NET, NS-1283.AWSDNS-32.ORG, NS-1707.AWSDNS-21.CO.UK, "
            "NS-421.AWSDNS-52.COM, NS-520.AWSDNS-01.NET",
            "2007-10-09T18:20:50Z",
            "2020-09-08T09:18:27Z",
            "2022-10-09T18:20:50Z",
        ]


class TestView03:
    def test_whois_panel(self, view03):
        whois = view03.whois_panel()
        assert type(whois) is Panel
        assert whois.title == Text("whois")

        # No Heading!

        # Table
        table = whois.renderable
        assert type(table) is Table
        assert table.columns[0].style == "bold bright_magenta"
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "Nameservers:",
            "Registered:",
            "Updated:",
            "Expires:",
        ]
        assert table.columns[1].style == "none"
        assert table.columns[1].justify == "left"
        assert [str(c) for c in table.columns[1]._cells] == [
            "dns1.bbc.com",
            "before Aug-1996",
            "10-Dec-2020",
            "13-Dec-2025",
        ]


class TestView04:
    def test_domain_panel(self, view04):
        domain = view04.domain_panel()
        assert type(domain) is Panel
        assert domain.title == Text("virustotal")

        # Heading
        assert domain.renderable.renderables[0] == Text(
            "google.com",
            spans=[Span(0, 10, 'bold yellow link https://virustotal.com/gui/domain/google.com')]
        )

        # Table
        table = domain.renderable.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == "bold bright_magenta"
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "Analysis:",
            "Reputation:",
            "Popularity:",
            "Last Modified:",
            "Last Seen:",
        ]
        assert table.columns[1].style == "none"
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                "1/94 malicious\nCyble",
                spans=[
                    Span(15, 20, ""),
                    Span(15, 20, "cyan"),
                ]
            ),
            Text("448"),
            Text(
                "Majestic (1)\nStatvoo (1)\nAlexa (1)\nCisco Umbrella (2)\nQuantcast (1)",
                spans=[
                    Span(10, 11, "cyan"),
                    Span(22, 23, "cyan"),
                    Span(32, 33, "cyan"),
                    Span(51, 52, "cyan"),
                    Span(65, 66, "cyan"),
                ]
            ),
            "2022-08-17T06:03:03Z",
            "2022-08-17T00:35:19Z",
        ]

    def test_resolutions_panel(self, view04):
        res = view04.resolutions_panel()
        assert res is None


class TestView05:
    def test_domain_panel(self, view05):
        domain = view05.domain_panel()
        assert type(domain) is Panel
        assert domain.title == Text("virustotal")

        # Heading
        assert domain.renderable.renderables[0] == Text(
            "tucows.com",
            spans=[Span(0, 10, 'bold yellow link https://virustotal.com/gui/domain/tucows.com')]
        )

        # Table
        table = domain.renderable.renderables[1]
        assert type(table) is Table
        assert table.columns[0].style == "bold bright_magenta"
        assert table.columns[0].justify == "left"
        assert table.columns[0]._cells == [
            "Analysis:",
            "Reputation:",
            "Last Modified:",
            "Last Seen:",
        ]
        assert table.columns[1].style == "none"
        assert table.columns[1].justify == "left"
        assert table.columns[1]._cells == [
            Text(
                "1/94 malicious\nDr.Web",
                spans=[
                    Span(15, 21, ""),
                    Span(15, 21, "cyan"),
                ]
            ),
            Text("-1"),
            "2022-08-17T05:30:23Z",
            "2022-08-16T22:24:18Z",
        ]


class TestView06:
    def test_whois_panel(self, view06):
        whois = view06.whois_panel()
        assert type(whois) is Panel
        assert whois.title == Text("whois")

        # Warning message
        assert whois.renderable == Text("Unable to gather whois data")



from pydantic import BaseModel, root_validator, validator
from typing import Any, Dict, List, Optional


class BaseData(BaseModel):
    attributes: Any
    id_: str
    type_: str

    class Config:
        fields = {
            "id_": "id",
            "type_": "type",
        }


class Meta(BaseModel):
    count: int


class AnalysisResult(BaseModel):
    category: str
    engine_name: str
    method: str
    result: str


class LastAnalysisResults(BaseModel):
    __root__: Dict[str, AnalysisResult]


class LastAnalysisStats(BaseModel):
    harmless: int
    malicious: int
    suspicious: int
    timeout: int
    undetected: int


class Popularity(BaseModel):
    rank: int
    timestamp: int


class PopularityRanks(BaseModel):
    __root__: Dict[str, Popularity]


class Attributes(BaseModel):
    creation_date: Optional[int]
    jarm: Optional[str]
    last_analysis_results: LastAnalysisResults
    last_analysis_stats: LastAnalysisStats
    last_dns_records_date: Optional[int]
    last_https_certificate_date: Optional[int]
    last_modification_date: Optional[int]
    last_update_date: Optional[int]
    popularity_ranks: PopularityRanks
    registrar: Optional[str]
    reputation: int
    tags: List[str]


class Data(BaseData):
    attributes: Attributes
    id_: str
    type_: str

    class Config:
        fields = {
            "id_": "id",
            "type_": "type",
        }


class Domain(BaseModel):
    data: Data


class ResolutionAttributes(BaseModel):
    date: int
    host_name: str
    resolver: str
    ip_address_last_analysis_stats: LastAnalysisStats
    ip_address: str
    host_name_last_analysis_stats: LastAnalysisStats


class ResolutionData(BaseData):
    attributes: ResolutionAttributes


class Resolutions(BaseModel):
    meta: Meta
    data: List[ResolutionData]


class HistoricalWhoisMap(BaseModel):
    domain: str = ""
    registrar: Optional[str]
    name_servers: List[str] = []
    creation_date: Optional[str]
    expiry_date: Optional[str]
    expiry_date_alt: Optional[str]
    admin_city: Optional[str]
    admin_state: Optional[str]
    admin_country: Optional[str]
    admin_postal_code: Optional[str]
    admin_email: Optional[str]
    last_updated: Optional[str]
    registered_on: Optional[str]
    registrant_email: Optional[str]
    registrant_name: Optional[str]
    registrant_org: Optional[str]
    updated_date: Optional[str]

    class Config:
        fields = {
            "domain": "Domain Name",
            "registrar": "Registrar",
            "name_servers": "Name Server",
            "creation_date": "Creation Date",
            "expiry_date": "Registry Expiry Date",
            "expiry_date_alt": "Expiry date",
            "admin_city": "Admin City",
            "admin_state": "Admin State/Province",
            "admin_country": "Admin Country",
            "admin_postal_code": "Admin Postal Code",
            "admin_email": "Admin Email",
            "last_updated": "Last updated",
            "registered_on": "Registered on",
            "registrant_email": "Registrant Email",
            "registrant_name": "Registrant Name",
            "registrant_org": "Registrant Organization",
            "updated_date": "Updated Date",
        }

    @validator("name_servers", pre=True)
    def transform_nameservers(cls, v):
        return v.split(" | ")

    @validator("*")
    def dedupe_values(cls, v):
        if "|" in v:
            return v.split(" | ")[0]
        return v


class HistoricalWhoisAttributes(BaseModel):
    first_seen_date: Optional[int]
    whois_map: Optional[HistoricalWhoisMap]
    registrant_country: Optional[str]
    registrar_name: Optional[str]
    last_updated: Optional[int]

    @root_validator(pre=True)
    def remove_empty_whois_map(cls, v):
        if v.get("whois_map") == {}:
            v.pop("whois_map")
        return v


class HistoricalWhoisData(BaseData):
    attributes: HistoricalWhoisAttributes


class HistoricalWhois(BaseModel):
    meta: Meta
    data: List[HistoricalWhoisData]



from wtfis.clients.base import BaseClient
from wtfis.models.ipwhois import IpWhois
from wtfis.models.virustotal import Resolutions

from typing import List, Optional


class IpWhoisClient(BaseClient):
    """
    IPWhois client
    """
    baseurl = "https://ipwho.is"

    def get_ipwhois(self, ip: str) -> Optional[IpWhois]:
        result = self._get(f"/{ip}")
        return IpWhois.parse_obj(result) if result.get("success") is True else None

    def bulk_get_ipwhois(
        self,
        resolutions: Resolutions,
        max_ips_to_enrich: int
    ) -> List[IpWhois]:
        results = []
        for idx, ip in enumerate(resolutions.data):
            if idx == max_ips_to_enrich:
                break
            ipwhois = self.get_ipwhois(ip.attributes.ip_address)
            if ipwhois:
                results.append(ipwhois)
        return results



class Theme:
    panel_title = "bright_blue"
    heading = "bold yellow"
    table_field = "bold bright_magenta"
    table_value = "none"
    inline_stat = "cyan"
    vendor_list = "cyan"
    nameserver_list = "cyan"
    disclaimer = "italic white on red"
    footer = "cyan"
    info = "bold green"
    warn = "bold yellow"
    error = "bold red"



from pydantic import BaseModel
from typing import List, Optional


class Registrant(BaseModel):
    organization: Optional[str]
    email: Optional[str]
    name: Optional[str]
    telephone: Optional[str]
    street: Optional[str]
    city: Optional[str]
    state: Optional[str]
    country: Optional[str]
    postalCode: Optional[str]


class Whois(BaseModel):
    contactEmail: Optional[str]
    domain: str
    expiresAt: Optional[str]
    lastLoadedAt: str
    name: str
    nameServers: List[str]
    organization: Optional[str]
    registered: Optional[str]
    registrant: Registrant
    registrar: Optional[str]
    registryUpdatedAt: Optional[str]



from pydantic import BaseModel


class Flag(BaseModel):
    img: str
    emoji: str
    emoji_unicode: str


class Connection(BaseModel):
    asn: int
    org: str
    isp: str
    domain: str


class IpWhois(BaseModel):
    ip: str
    success: bool
    type_: str
    continent: str
    continent_code: str
    country: str
    country_code: str
    region: str
    region_code: str
    city: str
    is_eu: bool
    postal: str
    calling_code: str
    capital: str
    borders: str
    flag: Flag
    connection: Connection

    class Config:
        fields = {
            "type_": "type",
        }
